/*
 * Copyright (c) 2025. Sergey S. Chernov - All Rights Reserved
 *
 *  You may use, distribute and modify this code under the
 * terms of the private license, which you must obtain from the author
 *
 * To obtain the license, contact the author: https://t.me/real_sergeych or email to
 * real dot sergeych at gmail.
 */

package net.sergeych.kiloparsec

import kotlinx.coroutines.sync.Mutex
import kotlinx.coroutines.sync.withLock

/**
 * Multiplatform atomically mutable value to be used in [kotlinx.coroutines],
 * with suspending mutating operations, see [mutate].
 *
 * Actual value can be either changed in a block of [mutate] when
 * new value _depends on the current value_ or with [reset].
 *
 * [value] getter is suspended because it waits until the mutation finishes
 */
open class AtomicAsyncValue<T>(initialValue: T) {
    private var actualValue = initialValue
    private val access = Mutex()

    /**
     * Change the value: get the current and set to the returned, all in the
     * atomic suspend operation. All other mutating requests including assigning to [value]
     * will be blocked and queued.
     * @return result of the mutation. Note that immediate call to property [value]
     *      could already return modified bu some other thread value!
     */
    suspend fun mutate(mutator: suspend (T) -> T): T = access.withLock {
        actualValue = mutator(actualValue)
        actualValue
    }

    /**
     * Atomic get or set the value. Atomic get means if there is a [mutate] in progress
     * it will wait until the mutation finishes and then return the correct result.
     */
    suspend fun value() = access.withLock { actualValue }

    /**
     * Set the new value without checking it. Shortcut to
     * ```mutate { value = newValue }```
     */
    suspend fun reset(value: T) = mutate { value }
}