/*
 * Copyright (c) 2025. Sergey S. Chernov - All Rights Reserved
 *
 *  You may use, distribute and modify this code under the
 * terms of the private license, which you must obtain from the author
 *
 * To obtain the license, contact the author: https://t.me/real_sergeych or email to
 * real dot sergeych at gmail.
 */

package net.sergeych.kiloparsec.adapter

import io.ktor.server.application.*
import io.ktor.server.routing.*
import io.ktor.server.websocket.*
import io.ktor.websocket.*
import kotlinx.coroutines.cancel
import kotlinx.coroutines.channels.Channel
import kotlinx.coroutines.channels.ClosedReceiveChannelException
import kotlinx.coroutines.isActive
import kotlinx.coroutines.launch
import net.sergeych.crypto2.SigningKey
import net.sergeych.kiloparsec.KiloInterface
import net.sergeych.kiloparsec.KiloServerConnection
import net.sergeych.kiloparsec.RemoteInterface
import net.sergeych.mp_logger.*
import net.sergeych.tools.AtomicCounter
import kotlin.time.Duration.Companion.seconds

/**
 * Create a ktor-based websocket server.
 * This call install Routing and WebSockets with proper configuration.
 *
 * The course of action is:
 *
 * - create LocalInterface and populate it with functionality
 * - call this method with localInterface
 * - optionally, connect the same interface with TCP or UDP providers on supported platforms,
 * in which case it might be useful to hae session creating function [createSession] separate.
 *
 * _Note_: [KiloInterface] as for now does not contain session creation in it as we suggest
 * session could be transport specific.
 *
 * @param localInterface where the actual work is performed.
 * @param path default http path to the websocket.
 * @param serverKey optional key to authenticate the connection. If the client specify expected
 *      server key it should match of connection will not be established.
 * @param createSession function to create a server session.
 */
fun <S> Application.setupWebsocketServer(
    localInterface: KiloInterface<S>,
    path: String = "/kp",
    serverKey: SigningKey? = null,
    createSession: () -> S,
) {
    install(WebSockets) {
        pingPeriod = 60.seconds //Duration.ofSeconds(15)
        timeout = 45.seconds
        maxFrameSize = Long.MAX_VALUE
        masking = false
    }
    val counter = AtomicCounter()
    routing {
        webSocket(path) {
            val log = LogTag("KWS:${counter.incrementAndGet()}")
            log.debug { "opening the connection" }
            val input = Channel<UByteArray>(256)
            val output = Channel<UByteArray>(256)
            launch {
                log.debug { "starting output pump" }
                while (isActive) {
                    try {
                        send(output.receive().toByteArray())
                    }
                    catch(_: ClosedReceiveChannelException) {
                        log.debug { "closing output pump as output channel is closed" }
                        break
                    }
                }
                log.debug { "closing output pump" }
            }
            val server = KiloServerConnection(
                localInterface,
                ProxyDevice(input, output),
                createSession(),
                serverKey
            )
            launch {
                server.run()
                close()
            }
            log.debug { "KSC started, looking for incoming frames" }
            for (f in incoming) {
                if (f is Frame.Binary)
                    try {
                        input.send(f.readBytes().toUByteArray())
                    } catch (_: RemoteInterface.ClosedException) {
                        log.warning { "caught local closed exception (strange!), closing" }
                        break
                    } catch (_: ClosedReceiveChannelException) {
                        log.info { "receive channel is closed, closing connection" }
                        break
                    } catch (t: Throwable) {
                        log.exception { "unexpected exception, server connection will close" to t }
                        break
                    }
                else
                    log.warning { "unknown frame type ${f.frameType}, ignoring" }
            }
            log.debug { "closing the server" }
            close()
            cancel()
            log.debug { "server wbesock processing done" }
        }
    }
}
