/*
 * Copyright (c) 2025. Sergey S. Chernov - All Rights Reserved
 *
 *  You may use, distribute and modify this code under the
 * terms of the private license, which you must obtain from the author
 *
 * To obtain the license, contact the author: https://t.me/real_sergeych or email to
 * real dot sergeych at gmail.
 */

package net.sergeych.kiloparsec

import assertThrows
import io.ktor.server.engine.*
import io.ktor.server.netty.*
import kotlinx.coroutines.delay
import kotlinx.coroutines.launch
import kotlinx.coroutines.runBlocking
import kotlinx.coroutines.test.runTest
import net.sergeych.crypto2.SigningSecretKey
import net.sergeych.crypto2.initCrypto
import net.sergeych.kiloparsec.adapter.setupWebsocketServer
import net.sergeych.kiloparsec.adapter.websocketClient
import net.sergeych.kiloparsec.adapter.websocketTransportDevice
import net.sergeych.mp_logger.Log
import java.net.InetAddress
import kotlin.random.Random
import kotlin.test.Test
import kotlin.test.assertEquals
import kotlin.test.assertFalse
import kotlin.test.assertTrue

class ClientTest {

    @Test
    fun testAddresses() {
        println(InetAddress.getLocalHost())
        println(InetAddress.getByName("localhost"))
        println(InetAddress.getByName("127.0.0.1"))
        println(InetAddress.getByName("mail.ru"))
    }


    @Test
    fun webSocketTest() = runTest {
        initCrypto()
//        fun Application.
        val cmdClose by command<Unit, Unit>()
        val cmdGetFoo by command<Unit, String>()
        val cmdSetFoo by command<String, Unit>()
        val cmdCheckConnected by command<Unit, Boolean>()

        Log.connectConsole(Log.Level.DEBUG)

        data class Session(var foo: String = "not set")

        var closeCounter = 0
        val serverInterface = KiloInterface<Session>().apply {
            var connectedCalled = false
            onConnected { connectedCalled = true }
            on(cmdGetFoo) { session.foo }
            on(cmdSetFoo) { session.foo = it }
            on(cmdCheckConnected) { connectedCalled }
            on(cmdClose) {
                throw LocalInterface.BreakConnectionException()
            }
        }

        val port = Random.nextInt(8080, 9090)
        val ns = embeddedServer(Netty, port = port, host = "0.0.0.0", module = {
            setupWebsocketServer(serverInterface) { Session() }
        }).start(wait = false)

        val client = websocketClient<Unit>("ws://localhost:$port/kp")
        val states = mutableListOf<Boolean>()
        val collector = launch {
            client.connectedStateFlow.collect {
                println("got: $closeCounter/$it")
                states += it
                if (!it) {
                    closeCounter++
                }
            }
        }
        assertEquals(true, client.call(cmdCheckConnected))
        assertTrue { client.connectedStateFlow.value }
        assertEquals("not set", client.call(cmdGetFoo))
        client.call(cmdSetFoo, "foo")
        assertEquals("foo", client.call(cmdGetFoo))

        assertTrue { client.connectedStateFlow.value }
        assertThrows<RemoteInterface.ClosedException> {
            client.call(cmdClose)
        }

        // connection should now be closed
        // the problem is: it needs some unspecified time to close
        // as it is async process.
        delay(100)
        assertFalse { client.connectedStateFlow.value }

        // this should be run on automatically reopen connection
        client.call(cmdSetFoo, "superbar")
        assertTrue { client.connectedStateFlow.value }
        assertEquals("superbar", client.call(cmdGetFoo))
        client.close()
        ns.stop()
        collector.cancel()
//        println("----= states: $states")
//        println("stopped server")
//        println("closed client")
    }

    @Test
    fun webSocketWaitForConnectTest() = runBlocking {
        initCrypto()
//        fun Application.
        Log.connectConsole(Log.Level.DEBUG)

        val clientInterface = KiloInterface<Unit>().apply {}

        val port = Random.nextInt(8080, 9090)
        var clientConnectCalls = 0
        // It should repeatedly reconnect, and we will count:
        KiloClient(clientInterface, SigningSecretKey.new()) {
            clientConnectCalls++
            KiloConnectionData(websocketTransportDevice("ws://localhost:$port/kp"), Unit)
        }
        delay(1200)
        // and check:
//        println("connection attemtps: $clientConnectCalls")
        assertTrue { clientConnectCalls > 1 }
    }

}