/*
 * Copyright (c) 2025. Sergey S. Chernov - All Rights Reserved
 *
 *  You may use, distribute and modify this code under the
 * terms of the private license, which you must obtain from the author
 *
 * To obtain the license, contact the author: https://t.me/real_sergeych or email to
 * real dot sergeych at gmail.
 */

package net.sergeych.kiloparsec.adapter

import io.ktor.network.sockets.*
import io.ktor.utils.io.core.*
import kotlinx.io.readByteArray
import net.sergeych.crypto2.toDump
import net.sergeych.kiloparsec.adapter.UdpBlock.Companion.CANCEL_BLOCK
import net.sergeych.kiloparsec.adapter.UdpBlock.Companion.ESCAPE_BYTE
import net.sergeych.kiloparsec.adapter.UdpBlock.Companion.PING_BLOCK
import net.sergeych.kiloparsec.adapter.UdpBlock.Companion.decode

/**
 * Encoded block for UDP datagram space-savvy. Minimum dara size is two bytes, which is fine
 * for Kiloparsec blocks.
 *
 * First byte is encoded using [ESCAPE_BYTE] depending on the second byte:
 *
 * | 0 | 1 | meaning |
 * |---|---|---------|
 * | [ESCAPE_BYTE] | [ESCAPE_BYTE] | Data block, dropping first byte |
 * | [ESCAPE_BYTE] | [PING_BLOCK] | Ping block, reset timers |
 * | [ESCAPE_BYTE] | [CANCEL_BLOCK] | close connection |
 * | any other | * | data block, all bytes |
 *
 * Use [encoded] and [toDatagram] to create binary or the datagram from a block, and [decode] to restore.
 *
 * We do not use serialization to speed up the transport layer.
 */
sealed class UdpBlock {
    /**
     * Block to show that the connection is closed and should also be closed on the other side
     */
    object Cancel : UdpBlock()

    /**
     * Parties show pings if there is no activity to keep it alive, detect connection loss and in some
     * cases revive NAT/Proxy state in routers.
     */
    object Ping : UdpBlock()

    /**
     * Parsec data block. Could not be smaller than two bytes.
     */
    class Data(val data: UByteArray) : UdpBlock() {
        override fun toString(): String {
            return "UDP Data (${data.size}):\n${data.toDump()}"
        }
        init {
            if( data.size < 2) throw IllegalArgumentException("data must be at least 2 bytes")
        }
    }

    val encoded: UByteArray by lazy {
        when(this) {
            is Data -> {
                // Do we need escaping?
                if( data[0] == ESCAPE_BYTE )
                    escapeAsArray + data
                else
                    data
            }
            is Cancel -> cancelAsArray
            is Ping -> pingAsArray
        }
    }

    fun toDatagram(address: SocketAddress): Datagram {
        val encoded = encoded.toByteArray()
        return Datagram(ByteReadPacket(encoded, 0, encoded.size), address)
    }

    companion object {
        val ESCAPE_BYTE = 255.toUByte()
        val PING_BLOCK = 0.toUByte()
        val CANCEL_BLOCK = 1.toUByte()

        private val escapeAsArray = ubyteArrayOf(ESCAPE_BYTE)
        private val pingAsArray = ubyteArrayOf(ESCAPE_BYTE, PING_BLOCK)
        private val cancelAsArray = ubyteArrayOf(ESCAPE_BYTE, CANCEL_BLOCK)

        fun decode(data: UByteArray): UdpBlock {
            if (data.size < 2)
                throw UdpTransportException("block too short: ${data.size}")
            return if( data[0] != ESCAPE_BYTE )
                // plain data
                Data(data)
            else {
                when(val b2 = data[1]) {
                    ESCAPE_BYTE -> {
                        // Escaped first byte, then plain data
                        Data(data.sliceArray(1 ..< data.size))
                    }
                    PING_BLOCK -> Ping
                    CANCEL_BLOCK -> Cancel
                    else -> throw UdpTransportException("invalid block type: $b2")
                }
            }
        }

        fun decode(datagram: Datagram) =
            decode(datagram.packet.readByteArray().toUByteArray())
    }
}