/*
 * Copyright (c) 2025. Sergey S. Chernov - All Rights Reserved
 *
 *  You may use, distribute and modify this code under the
 * terms of the private license, which you must obtain from the author
 *
 * To obtain the license, contact the author: https://t.me/real_sergeych or email to
 * real dot sergeych at gmail.
 */

package net.sergeych.kiloparsec.adapter

import io.ktor.network.selector.*
import io.ktor.network.sockets.*
import io.ktor.utils.io.CancellationException
import kotlinx.coroutines.*
import kotlinx.coroutines.flow.Flow
import net.sergeych.kiloparsec.KiloClient
import net.sergeych.kiloparsec.KiloServer
import net.sergeych.kiloparsec.RemoteInterface
import net.sergeych.mp_tools.globalLaunch
import net.sergeych.tools.AtomicCounter
import kotlin.time.Duration
import kotlin.time.Duration.Companion.minutes

internal val udpCounter = AtomicCounter(0)

class UdpTransportException(override val message: String) : RemoteInterface.InvalidDataException(message)

/**
 * Listen for incoming UDP connections and provide transport flow for it. See also [UdpServer.transportFlow]
 * for another way to create a server. Use it with [KiloServer]:
 * ```kotlin
 * // Whatever server session data we might need:
 * data class Session(
 *     var data: String,
 * )
 *
 * // declare some commands (normally in a shared module):
 * val cmdSave by command<String, Unit>()
 * val cmdLoad by command<Unit, String>()
 * val cmdDrop by command<Unit, Unit>()
 * val cmdException by command<Unit, Unit>()
 *
 * // Interface using the session above, can be shared between many
 * // server types and instances (different ports and protocols):
 * val cli = KiloInterface<Session>().apply {
 *     onConnected { session.data = "start" }
 *     on(cmdSave) { session.data = it }
 *     on(cmdLoad) {
 *         session.data
 *     }
 *     on(cmdException) {
 *         throw TestException()
 *     }
 *     on(cmdDrop) {
 *         throw LocalInterface.BreakConnectionException()
 *     }
 * }
 * // Now create a server to accept incoming UDPs on our port:
 * val server = KiloServer(cli, acceptUdpDevice(port)) {
 *     // This initializes new session for each incoming command
 *     Session("unknown")
 * }
 * ```
 *
 * See [connectUdpDevice] for the client sample.
 *
 * When it is necessary to stop listening to some port, use [UdpServer] instead.
 *
 * @param port port to listen
 * @param localInterface string form local interface to listen
 * @param maxInactivityTimeout maximum silence time after which the connection is supposed to be lost.
 *          the module automatically issues pings on inactivity when there is no data often enough
 *          to maintain the connection open.
 */
fun acceptUdpDevice(
    port: Int,
    localInterface: String = "0.0.0.0",
    maxInactivityTimeout: Duration = 2.minutes,
): Flow<InetTransportDevice> =
    UdpServer(port, localInterface,maxInactivityTimeout).transportFlow

/**
 * Connect to UDP server (see [acceptUdpDevice] or [UdpServer]) and return a [InetTransportDevice] for it. It
 * should be used with [KiloClient] as connection provider:
 * ```kotlin
 * val client = KiloClient<Unit>() {
 *     connect { connectUdpDevice("localhost:$port") }
 * }
 * // now we can execute remote commands:
 * assertEquals("start", client.call(cmdLoad))
 * ```
 *
 * @param hostPort "host:port" string address of the remote UDP port to connect to
 * @param maxInactivityTimeout maximum silence time after which the connection is supposed to be lost.
 *          the module automatically issues pings on inactivity when there is no data often enough
 *          to maintain the connection open.
 */
suspend fun connectUdpDevice(
    hostPort: String,
    maxInactivityTimeout: Duration = 2.minutes,
) = connectUdpDevice(hostPort.toNetworkAddress(), maxInactivityTimeout)

/**
 * Connect to UDP server (see [acceptUdpDevice]) and return a [InetTransportDevice] for it. It
 * should be used with [KiloClient] as connection provider:
 * ```kotlin
 * val client = KiloClient<Unit>() {
 *     connect { connectUdpDevice("localhost:$port") }
 * }
 * // now we can execute remote commands:
 * assertEquals("start", client.call(cmdLoad))
 * ```
 * @param addr the network address where to connect to
 * @param maxInactivityTimeout maximum silence time after which the connection is supposed to be lost.
 *          the module automatically issues pings on inactivity when there is no data often enough
 *          to maintain the connection open.
 */
suspend fun connectUdpDevice(
    addr: NetworkAddress,
    maxInactivityTimeout: Duration = 2.minutes,
): InetTransportDevice {
    val selectorManager = SelectorManager(Dispatchers.IO)
    val remoteAddress = InetSocketAddress(addr.host, addr.port)

    val done = CompletableDeferred<Unit>()

    val socket = aSocket(selectorManager).udp().connect(remoteAddress)
    val transport = UdpSocketTransport(object : UdpConnector {
        override suspend fun sendBlock(block: UdpBlock, toAddress: SocketAddress) {
            socket.send(block.toDatagram(remoteAddress))
        }

        override suspend fun disconnectClient(address: SocketAddress) {
            done.complete(Unit)
        }

    }, remoteAddress, false, maxInactivityTimeout)

    globalLaunch {
        launch {
            while (isActive) {
                try {
                    transport.processIncoming(UdpBlock.decode(socket.receive()))
                } catch (_: CancellationException) {
                    break
                } catch (e: Exception) {
                    transport.close()
                    break
                }
            }
        }
        done.await()
    }

    return transport.transportDevice
}