/*
 * Copyright (c) 2025. Sergey S. Chernov - All Rights Reserved
 *
 *  You may use, distribute and modify this code under the
 * terms of the private license, which you must obtain from the author
 *
 * To obtain the license, contact the author: https://t.me/real_sergeych or email to
 * real dot sergeych at gmail.
 */

package net.sergeych.kiloparsec.adapter

import io.ktor.network.selector.*
import io.ktor.network.sockets.*
import io.ktor.utils.io.*
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.IO
import kotlinx.coroutines.channels.ClosedReceiveChannelException
import kotlinx.coroutines.flow.flow
import kotlinx.coroutines.sync.Mutex
import kotlinx.coroutines.sync.withLock
import net.sergeych.kiloparsec.KiloServer
import net.sergeych.mp_logger.LogTag
import net.sergeych.mp_logger.Loggable
import net.sergeych.mp_logger.debug
import net.sergeych.mp_logger.exception
import net.sergeych.mp_tools.globalDefer
import kotlin.time.Duration
import kotlin.time.Duration.Companion.minutes

/**
 * UDP server for kiloparsec. Unlike [acceptUdpDevice], it allow stopping listening
 * to the port when need with [close]. Use [transportFlow] with [KiloServer], here is the
 * basic sample:
 *
 * ```kotlin
 * val uServer = UdpServer(port)
 * KiloServer(cli, uServer.transportFlow()) {
 *     Session("unknown")
 * }
 *
 * // server is now active and accepts connections
 * // ...
 *
 * // close and stop listening to the port:
 * uServer.close()
 * ```
 *
 * See [acceptUdpDevice] for more information.
 *
 * @param port port to listen to
 * @param localInterface string form of local interface to listen to
 * @param maxInactivityTimeout maximum silence time after which the connection is supposed to be lost.
 *          the module automatically issues pings on inactivity when there is no data often enough
 *          to maintain the connection open.

 */
class UdpServer(val port: Int, localInterface: String = "0.0.0.0", maxInactivityTimeout: Duration = 2.minutes) :
    Loggable by LogTag("UDPS${udpCounter.incrementAndGet()}"), UdpConnector {

    private val sessions = mutableMapOf<SocketAddress, UdpSocketTransport>()
    private val access = Mutex()

    private val selectorManager = SelectorManager(Dispatchers.IO)
    private val serverSocket = globalDefer {
        aSocket(selectorManager).udp().bind(InetSocketAddress(localInterface, port))
    }

    override suspend fun disconnectClient(address: SocketAddress) {
        access.withLock { sessions.remove(address) }
    }

    /**
     * a transport flow of [InetTransportDevice] suitable to be used with [KiloServer], see [UdpServer] for the
     * usage sample.
     */
    val transportFlow by lazy {
        flow {
            while (true) {
                try {
                    val datagram = serverSocket.await().receive()
                    val block = UdpBlock.decode(datagram)
                    val remoteAddress = datagram.address

                    access.withLock {
                        if (block == UdpBlock.Cancel) {
                            // if the cancel comes to already closed transport, do nothing
                            sessions.remove(remoteAddress)?.processIncoming(block)
                        } else {
                            sessions.getOrPut(remoteAddress) {
                                // new connection: create transport
                                debug { "Creating new connection to $remoteAddress" }
                                UdpSocketTransport(this@UdpServer, remoteAddress, true, maxInactivityTimeout)
                                    // and emit it:
                                    .also { emit(it.transportDevice) }
                            }.processIncoming(block)
                        }
                    }
                } catch (_: CancellationException) {
                    break
                } catch (_: ClosedReceiveChannelException) {
                    break
                } catch (e: Exception) {
                    exception { "unexpected exception in incoming datagram processing" to e }
                    close()
                    break
                }
            }
        }
    }

    override suspend fun sendBlock(block: UdpBlock, toAddress: SocketAddress) {
        serverSocket.await().send(block.toDatagram(toAddress))
    }

    suspend fun isClosed(): Boolean = serverSocket.await().isClosed

    /**
     * Close the UDP server. Calling it will cause:
     *
     * - Closing nound UDP socket on [port]
     * - Closing all pending connections
     * - cancelling the [transportFlow], which will cause Kiloparsec server to also stop
     *
     * Call suspends until socket and all sessions are closed. Later calls do nothing.
     */
    suspend fun close() {
        access.withLock {
            if (!isClosed()) {
                runCatching { serverSocket.await().close() }
            }
        }
        while (sessions.isNotEmpty()) {
            runCatching {
                access.withLock { sessions.values.firstOrNull() }
                    ?.close()
            }
        }
    }
}